(* PredomAux.v
   Part of the formalization for
   Biorthogonality, Step Indexing and Compiler Correctness
   N Benton & C-K Hur
   April 2009
*)


Require Import utility.
Require Import Predom.
(*Require Export Kleis. *)

Set Implicit Arguments.
Unset Strict Implicit.

Open Scope O_scope.

Implicit Arguments FST [D1 D2].
Implicit Arguments SND [D1 D2].
Implicit Arguments ID [D].
Implicit Arguments eta [D].
Implicit Arguments PAIR [D1 D2].
Implicit Arguments SWAP [D E].
Implicit Arguments EV [D E].
Implicit Arguments curry [D1 D2 D3].
Implicit Arguments Curry [D1 D2 D3].
Implicit Arguments KLEISLIR [A B C].
Implicit Arguments PROD_fun [A B C].
Implicit Arguments PROD_map [D1 D2 D3 D4].
Implicit Arguments Val [D].

(*==========================================================================
  Morphisms, missing from Predom.v
  ==========================================================================*)

Add Parametric Morphism (D E : cpo) : (@FST D E)
with signature (@Oeq (Dprod D E)) ==> (@Oeq D)
as PROD_FST_compat.
intros x y xy.
repeat (rewrite FST_simpl). repeat (rewrite Fst_simpl). 
auto. 
Defined.

Add Parametric Morphism (D E : cpo) : (@SND D E)
with signature (@Oeq (Dprod D E)) ==> (@Oeq E)
as PROD_SND_compat.
intros x y xy.
repeat (rewrite SND_simpl). repeat (rewrite Snd_simpl). 
auto. 
Defined.

Add Parametric Morphism (D E F:cpo) : (@PROD_fun E F D)
with signature (@Oeq (D -c> E)) ==> (@Oeq (D -c> F)) ==> (@Oeq (D -c> Dprod E F))
as PROD_fun_eq_compat_aux.
intros f0 f1 feq g0 g1 geq.
refine (fcont_eq_intro _).
intro d.
assert (gg:=fcont_eq_elim geq d). assert (ff := fcont_eq_elim feq d).
auto.
Qed.

Add Parametric Morphism (D E F G H:cpo) : (@fcont_comp3 D E F G H)
with signature (@Oeq (E -c> F -C-> G -C-> H)) ==> (@Oeq (D -c> E)) ==> (@Oeq (D -c> F)) ==> (@Oeq (D -c> G)) ==> (@Oeq (D -c> H))
as fcont_comp3_eq_compat_aux.
intros h0 h1 heq e0 e1 eeq f0 f1 feq g0 g1 geq.
simpl. apply fcont_eq_intro. intros d.
repeat (rewrite fcont_comp3_simpl).
refine (app_eq_compat _ _).
refine (app_eq_compat _ _).
refine (app_eq_compat _ _).
apply heq.
apply (fcont_eq_elim eeq d).
apply (fcont_eq_elim feq d).
apply (fcont_eq_elim geq d).
Qed.

Add Parametric Morphism D E F : (@curry D E F)
with signature (@Oeq (Dprod D E -c> F)) ==> (@Oeq (D -c> E -C-> F))
as Curry_eq_compat_aux.
intros f0 f1 feq. refine (fcont_eq_intro _). intros d. refine (fcont_eq_intro _). intros e.
repeat (rewrite curry_simpl). auto.
Qed.

Add Parametric Morphism (D E F G:cpo) : (@PROD_map D E F G)
with signature (@Oeq (D -c> F)) ==> (@Oeq (E -c> G)) ==> (@Oeq (@Dprod D E -c> @Dprod F G))
as PROD_map_eq_compat.
intros f g feq s r seq.
simpl. apply fcont_eq_intro.
intros x. case x. clear x. intros d e.
assert (fd:=fcont_eq_elim feq d). assert (fe := fcont_eq_elim seq e).
auto.
Qed.



(*==========================================================================
  Some basic lemmas
  ==========================================================================*)

Lemma FST_SWAP : forall D E F (f:D -c> F), f @_ FST @_ SWAP == f @_ SND (D1:=E).
intros D E F f.
apply fcont_eq_intro. auto. 
Qed.

Lemma SND_SWAP : forall D E F (f:E -c> F), f @_ SND @_ SWAP == f @_ FST (D2:=D).
intros D E F f.
apply fcont_eq_intro. auto. 
Qed.

Lemma SND_PAIR : forall D E F (f:E -c> F) x, f @_ SND @_ PAIR (D1:=D) x == f. 
intros D E F f x.
apply fcont_eq_intro. auto. 
Qed.

Lemma FST_PAIR : forall D E F (f:D -c> F) (x:D),  f @_ FST @_ PAIR x == K E (f x). 
intros D E F f x.
apply fcont_eq_intro. auto. 
Qed.

Lemma FST_PROD_fun_aux : forall C D E F (f:C -c> D) (g:C -c> E) (k:D -c> F), k @_ FST @_ PROD_fun f g == k @_ f.
intros C D E F f g k.
apply fcont_eq_intro. auto.  
Qed.

Lemma SND_PROD_fun_aux : forall C D E F (f:C -c> D) (g:C -c> E) (k:E -c> F), k @_ SND @_ PROD_fun f g == k @_ g.
intros C D E F f g k.
apply fcont_eq_intro. auto.
Qed.

Lemma FST_PROD_fun : forall C D E (f:C -c> D) (g:C -c> E), FST @_ PROD_fun f g == f.
intros C D E f g.
apply fcont_eq_intro. auto.  
Qed.

Lemma SND_PROD_fun : forall C D E (f:C -c> D) (g:C -c> E), SND @_ PROD_fun f g == g.
intros C D E f g.
apply fcont_eq_intro. auto.
Qed.

Lemma PROD_fun_comp : 
  forall C D E F (f:D-c>E) (g:D-c>F) (h:C-c>D), PROD_fun f g @_ h == PROD_fun (f @_ h) (g @_ h).
intros C D E F f g h.
apply fcont_eq_intro. auto.
Qed.
 
Lemma ID_comp_left :
  forall C D (f:C-c>D), ID @_ f == f.
intros C D f.
apply fcont_eq_intro. auto. 
Qed.

Lemma ID_comp_right :
  forall C D (f:C-c>D), f @_ ID == f.
intros C D f.
apply fcont_eq_intro. auto. 
Qed.

Lemma PROD_map_ID_ID :
  forall C D, PROD_map (D1 := C) (D2 := D) ID ID == ID.
intros C D. 
apply fcont_eq_intro. auto. 
Qed.

Lemma eta_injective :
  forall D E (f g:D -c> E), eta @_ f == eta @_ g -> f == g.
intros D E f g eq.
refine (fcont_eq_intro _).
intro x. 
assert (H := fcont_eq_elim eq x).
repeat (rewrite fcont_comp_simpl in H).
repeat (rewrite eta_val in H).
exact (vinj H).
Qed.

Lemma PROD_map_PROD_map : forall A B C D E F (f:A -c> B) (g : C -c> D) (h :B -c> E) (i :D -c> F), 
  PROD_map h i @_ PROD_map f g == PROD_map (h @_ f) (i @_ g). 
intros. apply fcont_eq_intro.  auto. 
Qed.

Lemma Curry_EV : forall C D E (f:C -c> D -C-> E) , curry (EV @_ PROD_map f ID) == f.
intros C D E f.
symmetry.
apply curry_unique. trivial.
Qed.

Lemma EV_Curry : forall C D E (f:Dprod C D -c> E) , EV @_ PROD_map (curry f) ID == f.
intros C D E f.
rewrite <- curry_com.  trivial.
Qed. 

Lemma Curry_comp : forall A B C D (f:Dprod A B -c> C) (g:D -c> A), curry f @_ g == curry (f @_ PROD_map g ID).
intros A B C D f g.
assert (H := curry_com f). rewrite H. 
rewrite fcont_comp_assoc.
rewrite PROD_map_PROD_map. 
rewrite Curry_EV.  rewrite ID_comp_left.
rewrite Curry_EV. trivial.
Qed.

Lemma KLEISLIR_ValVal : forall (D E F : cpo) (e1: D -C-> DL E) (e2 : Dprod D E -C-> (DL F)) env r,
                (KLEISLIR e2 @_ PROD_fun ID e1) env == Val r -> exists d, e1 env == Val d /\ e2 (PAIR env d) == Val r.
intros D E F e1 e2 env r eq.
destruct eq as [eq1 eq2].
destruct (DLvalval eq2) as [n [d [pd ld]]].
admit.
Qed.

