
(****
   Examples of observational equivalences for
   Biorthogonality, Step-Indexing & Compiler Correctness
   Nick Benton & Chung-Kil Hur
   April 2009
 ****)

Require Import Program.Equality.
Require Import Program.Tactics.
Require Import FunctionalExtensionality.
Require Import utility.
Require Import domoprs.
Require Import toptopsem_dom.
Require Import Predom.
Require Import typedlambda.
Require Import typeddensem.
Require Import compiler_correct.
Set Implicit Arguments.
Unset Strict Implicit.


(***
   Note that we use the classical logic to reason about examples.
***)
Require Import Classical.

Lemma DL_bot_dec: forall D (d:DL D), (d == @DL_bot D) \/ (exists d0, d == Val d0).
intros.
destruct (classic (forall n, exists d', pred_nth d n = Eps d')).

left.
split. simpl.
apply (@DLle_rec D (fun (x: DL D) => fun (y: DL D) =>
  (forall n, exists x', pred_nth x n = Eps x') /\
  (forall n, exists y', pred_nth y n = Eps y'))).

intros. destruct H0. split.
intros. destruct (H0 (S n)). exists x0. auto.
intros. destruct (H1 (S n)). exists x0. auto.

intros. destruct H0. split.
intros. destruct (H0 (S n)). exists x0. auto.
auto.

intros. destruct H0. destruct (H0 0). inversion H2.

split. intros. apply H.
intros. exists (DL_bot D).
induction n.
rewrite DL_bot_eq at 1. auto.
rewrite DL_bot_eq at 1. apply IHn.

apply (@DL_bot_least D).

right.
hreplace H with (@not_all_ex_not _ _ H). destruct H.
assert (exists dd, pred_nth d x = dd). exists (pred_nth d x). auto.
destruct  H0. destruct x0.
absurd False; auto. apply H. exists x0. auto.
exists t.
apply Oeq_sym. eapply (@Val_exists_pred_eq D).
exists x. auto.
Qed.

Lemma cesd_val_eq_dec: forall (v1 v2:ValueM), 
  v1 = v2 \/ v1 <> v2.
intros. apply classic.
Qed.

Lemma Smash_unfold: forall C D (x:DL C) (y:DL D),
  Smash C D x y == 
  kleisli_cc (D -C-> DL (Dprod C D)) (Dprod C D)
  (kleisli_C D (Dprod C D) y)
  (kleisli_cc C (D -C-> DL (Dprod C D)) (eta @_ curry eta) x).

intros.
unfold Smash, Operator2.
rewrite fcont_comp_simpl.
unfold kleisli_cc.
rewrite curry_simpl.
rewrite fcont_comp_simpl.
unfold Application.
rewrite curry_simpl.
unfold SWAP.
rewrite PROD_fun_simpl, PAIR_simpl, Pair_simpl.
rewrite FST_simpl, Fst_simpl, SND_simpl, Snd_simpl.
simpl fst. simpl snd.
rewrite uncurry_simpl.
rewrite fcont_comp2_simpl.
repeat rewrite kleisli_C_simpl.
rewrite fcont_comp_simpl.
rewrite SND_simpl, Snd_simpl. simpl snd.
repeat rewrite <- kleisli_cc_simpl.
rewrite FST_simpl, Fst_simpl. simpl fst.
auto.
Qed.

Lemma Smash_lval: forall (C D:cpo) (x: C) (y:DL D),
  Smash C D (Val x) y == kleisli_cc _ _ (curry eta x) y.

intros.
rewrite Smash_unfold.
pattern kleisli_cc at 1.
rewrite kleisli_cc_simpl, kleisli_inv, fcont_comp_simpl, eta_val.
rewrite kleisli_cc_simpl, kleisli_inv.
auto.
Qed.

(*
Lemma Smash_rval: forall (C D:cpo) (x: DL C) (y: D),
  Smash C D x (Val y) == @kleisli_cc _ _ (curry (eta @_ SWAP) y) x.
*)

Lemma Smash_lbot: forall (C D:cpo) (y: DL D),
  Smash C D (DL_bot _) y == DL_bot _.
intros.
rewrite Smash_unfold.
pattern kleisli_cc at 1.
repeat rewrite kleisli_cc_simpl, kleisli_bot.
auto.
Qed.

Lemma Smash_rbot: forall (C D:cpo) (x: DL C),
  Smash C D x (DL_bot _) == DL_bot _.

intros.
rewrite Smash_unfold.

destruct (DL_bot_dec x).
rewrite H.
pattern kleisli_cc at 1.
rewrite kleisli_cc_simpl, kleisli_bot.
rewrite kleisli_cc_simpl, kleisli_bot.
auto.

destruct H.
rewrite H.
pattern kleisli_cc at 1.
rewrite kleisli_cc_simpl, kleisli_Val.
rewrite fcont_comp_simpl, eta_val.
rewrite kleisli_cc_simpl, kleisli_Val.
rewrite kleisli_C_simpl, kleisli_bot.
auto.
Qed.

(****
   Project Fst
****)

Definition proj_fst_src E A B : Value E (A-->B-->A) :=
  TLAM (TVAL (TLAM (TVAL (TVAR (SVAR _ (ZVAR _ _)))))).

Definition proj_fst_code E A B := compileV (proj_fst_src E A B).

Lemma proj_fst_simpl: forall E A B e,
  SemVal (proj_fst_src E A B) e == eta @_ (curry (eta @_ FST)).
intros.
unfold proj_fst_src, SemVal, SemVar. fold SemEnv.
apply (@fcont_eq_intro (SemType A) (DL (SemType (B-->A)))). intro.
repeat rewrite sem_op_simpl. unfold sem_curry.
rewrite CURRY_simpl, Curry_simpl.
repeat rewrite fcont_comp_simpl.
apply fcont_eq_compat.
apply (@fcont_eq_intro (SemType B) (DL (SemType A))). intro.
rewrite curry_simpl, CURRY_simpl, Curry_simpl.
repeat rewrite fcont_comp_simpl.
auto.
Qed.

Lemma proj_fst_sem_eq: forall E A B (dc1:SemCTypeL E A) (dc2: SemCTypeL E B),
  (forall de, exists v, dc2 de == Val v) -> 
  sem_app _ _ _ (sem_app _ _ _ (@eta _ @_ (SemVal (proj_fst_src E A B))) dc1) dc2 == dc1.
fold SemType; intros.

repeat rewrite sem_app_eval.
apply (@fcont_eq_intro (SemEnv E) (DL (SemType A))); intros. 
rename x into e.

repeat rewrite fcont_comp_simpl.
rewrite PROD_fun_simpl, PAIR_simpl, Pair_simpl.
rewrite uncurry_simpl.
repeat rewrite fcont_comp_simpl.

hreplace H with (H e).
destruct H.
rewrite H.
rewrite PROD_fun_simpl, PAIR_simpl, Pair_simpl.
repeat rewrite fcont_comp_simpl.
rewrite eta_val.
rewrite uncurry_simpl.

destruct (DL_bot_dec (dc1 e)).
(* dc1 e == DL_bot *)
rewrite H0.
rewrite Smash_rbot.
setoid_rewrite kleisli_cc_simpl at 2.
rewrite kleisli_bot.
rewrite Smash_lbot.
rewrite kleisli_cc_simpl, kleisli_bot.
auto.

(* dc1 e <> DL_bot *)
destruct H0.
rewrite H0.
unfold Smash. rewrite Operator2_simpl, eta_val.
setoid_rewrite kleisli_cc_simpl at 2.
rewrite kleisli_Val.
unfold EV.
rewrite fcont_comp2_simpl, AP_simpl.
rewrite FST_simpl, Fst_simpl, SND_simpl, Snd_simpl. unfold fst, snd.
rewrite proj_fst_simpl, fcont_comp_simpl, eta_val.
rewrite Operator2_simpl, eta_val.
rewrite kleisli_cc_simpl, kleisli_Val.
auto.
Qed.

Lemma proj_fst_eq: forall E A B c1 c2 (dc1:SemCTypeL E A) (dc2: SemCTypeL E B),
  c1 =P= dc1 -> c2 =P= dc2 -> 
  (forall de, exists v, dc2 de == Val v) -> 
  (ComposeApp (ComposeApp (proj_fst_code E A B) c1) c2) =P= dc1.

intros.
assert ((proj_fst_code E A B) =P= @eta _ @_ SemVal (proj_fst_src E A B)).
apply compiler_correctness.
assert (comprel := compositionality_funapp H2 H); fold SemType in comprel.
hreplace comprel with (compositionality_funapp comprel H0); fold SemType in comprel.

eapply relEqC_eq.
Focus 2.
apply comprel.
apply Oeq_sym.
apply (@proj_fst_sem_eq E A B dc1 dc2 H1).
Qed.

Corollary proj_fst_fun_obseq: forall E A B c1 c2 (dc1:SemCTypeL E A) (dc2: SemCTypeL E B),
  c1 =P= dc1 -> c2 =P= dc2 -> 
  (forall de, exists v, dc2 de == Val v) ->
  E |- (ComposeApp (ComposeApp (proj_fst_code E A B) c1) c2) ~P~ c1 ::: A .

intros.
exists dc1.
split.
eapply proj_fst_eq.
assumption.
apply H0.
assumption.
assumption.
Qed.


(****
   Commutativity of Plus 
*****)

Definition plus_src E : Value E (Int-->Int-->Int) :=
  TLAM (TVAL (TLAM (
    TOP PLUS 
    (TVAR (SVAR _ (ZVAR _ _)))
    (TVAR (ZVAR _ _))))).

Definition plus_code E := compileV (plus_src E).

Lemma plus_comm: forall E c1 c2 (dc1:SemCTypeL E Int) (dc2: SemCTypeL E Int),
  c1 =P= dc1 -> c2 =P= dc2 ->
  (ComposeApp (ComposeApp (plus_code E) c1) c2) =P= 
  (sem_app _ _ _ (sem_app _ _ _ (@eta _ @_ (SemVal (plus_src E))) dc1) dc2 : SemCTypeL E Int).

fold SemType; intros.

assert ((plus_code E) =P= (@eta _ @_ (SemVal (plus_src E)))).
apply compiler_correctness.
assert (ComposeApp (plus_code E) c1 =P= ((sem_app _ _ _ (@eta _ @_ (SemVal (plus_src E))) dc1) : SemCTypeL E (Int-->Int))).
apply (compositionality_funapp H1 H).
fold SemType in H2.
apply (compositionality_funapp H2 H0).
Qed.

Lemma plus_src_simpl: forall E e,
  SemVal (plus_src E) e == eta @_ (curry (eta @_ (uncurry _ _ _ (SimpleOp2 plus)))).

intros.
unfold plus_src, SemVal, SemVar.
apply (@fcont_eq_intro (SemType Int) (DL (SemType (Int-->Int)))). intro.
fold SemEnv. repeat rewrite sem_op_simpl. unfold sem_curry.
repeat rewrite fcont_comp_simpl.
repeat rewrite CURRY_simpl, Curry_simpl.
repeat rewrite fcont_comp_simpl.
apply fcont_eq_compat.
apply (@fcont_eq_intro (SemType Int) (DL (SemType Int))). intro.
rewrite curry_simpl, CURRY_simpl, Curry_simpl.
repeat rewrite fcont_comp_simpl.
apply fcont_eq_compat.
rewrite sem_op2_eval. auto.
Qed.

Lemma plus_src_simpl2: forall E e x,
  SemVal (plus_src E) e x == @Val (_-C->_) (eta @_ SimpleOp2 plus x).
intros.
rewrite plus_src_simpl.
repeat rewrite fcont_comp_simpl.
rewrite eta_val.
apply (@eqDLeq (SemType (Int-->Int))).

apply (@fcont_eq_intro (SemType Int) (DL (SemType Int))); intro.
rewrite curry_simpl.
repeat rewrite fcont_comp_simpl.
rewrite uncurry_simpl.
auto.
Qed.

Lemma sem_plus_comm : forall E d1 d2,
   sem_app _ _ _ (sem_app _ _ _ (eta @_ SemVal (plus_src E)) d1) d2 ==
   sem_app _ _ _ (sem_app _ _ _ (eta @_ SemVal (plus_src E)) d2) d1.

fold SemType; intros.

repeat rewrite sem_app_eval.
apply (@fcont_eq_intro (SemEnv E) (DL (SemType Int))); intro.

repeat rewrite fcont_comp_simpl.
repeat rewrite PROD_fun_simpl, PAIR_simpl, Pair_simpl.
repeat rewrite uncurry_simpl.
repeat rewrite fcont_comp_simpl.
repeat rewrite PROD_fun_simpl, PAIR_simpl, Pair_simpl. 
repeat rewrite fcont_comp_simpl, eta_val.
repeat rewrite uncurry_simpl.

destruct (DL_bot_dec (d1 x)).
(* d1 x = DL_bot *)
rewrite H. 
rewrite Smash_rbot.  
setoid_rewrite kleisli_cc_simpl at 2.
rewrite kleisli_bot. 
rewrite Smash_lbot.
rewrite Smash_rbot.
auto.

(* d1 x <> DL_bot *)
destruct H.
rewrite H. 
unfold Smash at 2. rewrite Operator2_simpl, eta_val. 
setoid_rewrite kleisli_cc_simpl at 2.
rewrite kleisli_Val.

destruct (DL_bot_dec (d2 x)).
(* d2 x = DL_bot *)
rewrite H0. 
rewrite Smash_rbot.
rewrite kleisli_cc_simpl, kleisli_bot.
rewrite Smash_rbot.
setoid_rewrite kleisli_cc_simpl at 2. rewrite kleisli_bot.
rewrite Smash_lbot.
rewrite kleisli_cc_simpl, kleisli_bot.
auto.

(* d2 x <> DL_bot *)
destruct H0.
rewrite H0.
unfold Smash at 3. rewrite Operator2_simpl, eta_val.
setoid_rewrite kleisli_cc_simpl at 3; rewrite kleisli_Val.

unfold EV.
repeat rewrite fcont_comp2_simpl.
repeat rewrite FST_simpl, Fst_simpl, SND_simpl, Snd_simpl.
repeat rewrite AP_simpl. unfold fst, snd.
repeat rewrite plus_src_simpl2.
unfold Smash. repeat rewrite Operator2_simpl, eta_val.
repeat rewrite kleisli_cc_simpl, kleisli_Val.
repeat rewrite fcont_comp2_simpl, AP_simpl.
repeat rewrite FST_simpl, Fst_simpl, SND_simpl, Snd_simpl. simpl.
repeat rewrite fcont_comp_simpl.
repeat rewrite eta_val.
apply (@eqDLeq (SemType Int)).
repeat rewrite SimpleOp2_simpl.
assert (x0 + x1 = x1 + x0)%nat. omega.
rewrite H1. auto.
Qed.

Theorem plus_comm_obseq: forall E c1 c2 (dc1:SemCTypeL E Int) (dc2: SemCTypeL E Int) (de:SemEnv E),
  c1 =P= dc1 -> c2 =P= dc2 ->
    E |- (ComposeApp (ComposeApp (plus_code E) c1) c2) 
      ~P~ (ComposeApp (ComposeApp (plus_code E) c2) c1) ::: Int.

intros.
exists (sem_app _ _ _ (sem_app _ _ _ (@eta _ @_ SemVal (plus_src E)) dc1) dc2).
split.
apply plus_comm.
assumption.
assumption.

apply relEqC_eq with (sem_app _ _ _ (sem_app _ _ _ (eta @_ SemVal (plus_src E)) dc2) dc1).
apply sem_plus_comm.
apply plus_comm.
assumption.
assumption.
Qed.



(****
   Optimization example
****)


Definition id_src E T : Value E (T-->T) := (TLAM (TVAL (TVAR (ZVAR _ _)))).
Definition id_code E T := compileV (id_src E T).
Definition id_fun e := VC e [PushV 0, Ret].

Fixpoint iEnv (T:Ty) (E: Env) i {struct i} : Ty :=
  match i with
    | 0 => T
    | S i => match E with
               | nil => T
               | T'::E' =>  iEnv T' E' i
             end
  end.

Fixpoint VTmVar (T:Ty) (E:Env) i {struct i} : Var (T::E) (iEnv T E i) :=
  match i with
    | 0 => ZVAR _ _
    | S i => match E with
               | nil => ZVAR _ _
               | T'::E' => SVAR _ (VTmVar T' E' i)
             end
  end.

Definition VTm (T:Ty) (E:Env) i := TVAR (VTmVar T E i).
Implicit Arguments VTm [T E].

Program Definition applyn_fFnv E T :  
  Exp (T::Int::(Int-->T-->T)::(T-->T)::E) T :=

  TLET (TGT (VTm 1) (TINT 0)) 
    (TIF (VTm 0)
      (TLET
        (TLET (TOP (MINUS) (VTm 2) (TINT 1))
          (TAPP 
            (VTm 4) 
            (VTm 0)))
        (TLET
          (TAPP 
            (VTm 0)
            (VTm 2))
          (TAPP
            (VTm 6)
            (VTm 0)
          )))
      (TVAL (VTm 1))
    ).

Definition applyn_fFn E T : Value (Int::(Int-->T-->T)::T-->T::E) (T-->T) :=
  TLAM (applyn_fFnv E T).

Definition applyn_f E T : Value (T-->T::E) (Int-->T-->T) :=
  TFIX (TVAL (applyn_fFn E T)).

Definition applyn E T : Value E ((T-->T)-->Int-->T-->T) :=
  TLAM (TVAL (applyn_f E T)).

Definition applyn_code E T := compileV (applyn E T).

Definition applyn_opt E T:=
  [PushC [PushC [PushV 0,PushV 1,EqTest,
    Sel
    [PushC (id_code E T++[Ret]), Join] 
    ((applyn_code nil T)++[PushV 0, App, Join]),
  Ret], Ret]]++
  (id_code E T)++[App].

Fixpoint CtxExtRen E E' T (var: Var E T) : Var (E++E') T :=
  match var with
    | ZVAR _ _ => ZVAR _ _
    | SVAR _ _ _ var' => SVAR _ (CtxExtRen _ var')
  end.

Lemma liftRenaming_CtxExtRen : forall T E E', 
  liftRenaming T (@CtxExtRen E E') = @CtxExtRen (T::E) E'.
intros. apply MapExtensional. intros.
dependent induction var. auto. auto.
Qed.

Lemma CtxExt_preserve_sem:
  (forall E T (tv:Value E T) E' de de', projEnv E E' de' = de -> SemVal (renameVal (@CtxExtRen E E') tv) de' == SemVal tv de)
  /\
  (forall E T (tc:Exp E T) E' de de', projEnv E E' de' = de -> SemExp (renameExp (@CtxExtRen E E') tc) de' == SemExp tc de).

apply ExpVal_ind.

auto. auto.

intros. simpl. rewrite <- H.
induction v. auto.
simpl. repeat rewrite fcont_comp_simpl.
rewrite IHv. auto. reflexivity.

intros. rewrite renameTLAM. simpl. fold SemEnv.
apply fcont_eq_intro. intro.
unfold sem_curry. repeat rewrite CURRY_simpl, Curry_simpl. 
rewrite liftRenaming_CtxExtRen.
rewrite <- H. simpl. reflexivity.
rewrite <- H0. auto.

intros. rewrite renameTFIX. simpl.
apply fcont_eq_intro. intro.
unfold sem_fix. fold SemType SemEnv. 
repeat rewrite curry_simpl.
repeat rewrite fcont_comp_simpl.
repeat rewrite uncurry_simpl.
repeat rewrite fcont_comp_simpl.
repeat rewrite CURRY_simpl.
rewrite liftRenaming_CtxExtRen.
assert (TEMP:= @liftRenaming_CtxExtRen ty1 ((ty1-->ty2)::env) E'). simpl in TEMP. rewrite TEMP. clear TEMP.
apply fcont_eq_compat2.
apply (@fcont_eq_intro (SemType (ty1-->ty2)) (SemType (ty1 --> ty2))). intros.
apply (@fcont_eq_intro (SemType ty1) (DL (SemType ty2))). intros.
repeat rewrite Curry_simpl.
rewrite <- H.
simpl. reflexivity.
rewrite <- H0. auto.
auto.

intros. rewrite renameTPAIR. simpl.
repeat rewrite sem_op2_eval.
rewrite H. rewrite H0. reflexivity. auto. auto.

intros. rewrite renameTFST. simpl. fold SemEnv SemType.
repeat rewrite sem_op_simpl.
repeat rewrite fcont_comp_simpl.
rewrite H. reflexivity. auto.

intros. rewrite renameTSND. simpl. fold SemEnv SemType.
repeat rewrite sem_op_simpl.
repeat rewrite fcont_comp_simpl.
rewrite H. reflexivity. auto.

intros. rewrite renameTOP. simpl.
repeat rewrite sem_op_simpl.
repeat rewrite fcont_comp_simpl.
repeat rewrite sem_op2_eval.
rewrite H. rewrite H0. reflexivity.
auto. auto.

intros. rewrite renameTGT. simpl.
repeat rewrite sem_op_simpl.
repeat rewrite fcont_comp_simpl.
repeat rewrite sem_op2_eval.
rewrite H. rewrite H0. reflexivity.
auto. auto.

intros. rewrite renameTVAL. simpl.
repeat rewrite sem_op_simpl.
repeat rewrite fcont_comp_simpl.
repeat rewrite sem_op2_eval.
rewrite H. reflexivity.
auto.

intros. rewrite renameTLET. simpl. fold SemType.
repeat rewrite sem_op_simpl.
repeat rewrite sem_app_eval.
repeat rewrite fcont_comp_simpl.
apply (fcont_eq_compat (kleisli_cc (Dprod (SemType ty1 -C-> DL (SemType ty2)) (SemType ty1)) (SemType ty2) EV)).
apply fcont_eq_compat.
repeat rewrite PROD_fun_simpl.
repeat rewrite fcont_comp_simpl.
apply fcont_eq_compat2.
apply fcont_eq_compat.
apply (@fcont_eq_intro (SemType ty1) (DL (SemType ty2))). intro.
unfold sem_curry. repeat rewrite CURRY_simpl, Curry_simpl. 
rewrite liftRenaming_CtxExtRen.
rewrite <- H0. simpl. reflexivity. rewrite <- H1. auto.
auto.

intros. rewrite renameTAPP. simpl. fold SemType.
repeat rewrite sem_op_simpl.
repeat rewrite sem_app_eval.
repeat rewrite fcont_comp_simpl.
apply (fcont_eq_compat (kleisli_cc (Dprod (SemType ty1 -C-> DL (SemType ty2)) (SemType ty1)) (SemType ty2) EV)).
apply fcont_eq_compat.
repeat rewrite PROD_fun_simpl.
repeat rewrite fcont_comp_simpl.
rewrite H, H0. reflexivity.
auto. auto.

intros. rewrite renameTIF. simpl.
repeat rewrite sem_op3_eval.
rewrite H, H0, H1. reflexivity.
auto. auto. auto.
Qed.

Lemma applyn_CtxExt: forall E T, renameVal (@CtxExtRen nil E) (applyn nil T) = applyn E T.
auto.
Qed.

Lemma applyn_sem_simpl: forall E T de,
  SemVal (applyn E T) de == SemVal (applyn nil T) tt.
intros. rewrite <- applyn_CtxExt. 
apply (proj1 CtxExt_preserve_sem _ _ (applyn nil T)).
auto.
Qed.

Definition applyn_fFnv_reduce m T v c e e' s d :
  reduceSet (App :: c, e,
    v :: VC [VN m, VRC [VC e' [PushV 0, Ret]] (compileV (applyn_fFn nil T) ++ [Ret]), VC e' [PushV 0, Ret]]
    (compileE (applyn_fFnv nil T) ++ [Ret]) :: s, d) (c, e, v :: s, d).
induction m.
intros. simpl.
eapply reduceS. apply step_app.
eapply reduceS. apply step_pushc.
eapply reduceS. apply step_pushv. reflexivity.
eapply reduceS. apply step_pushn.
eapply reduceS. apply step_swap.
eapply reduceS. apply step_op.
eapply reduceS. apply step_app.
eapply reduceS. apply step_pushv. reflexivity.
eapply reduceS. apply step_selfalse.
eapply reduceS. apply step_pushv. reflexivity.
eapply reduceS. apply step_join.
eapply reduceS. apply step_ret.
eapply reduceS. apply step_ret.
apply reduce0.

intros. simpl.
eapply reduceS. apply step_app.
eapply reduceS. apply step_pushc.
eapply reduceS. apply step_pushv. reflexivity.
eapply reduceS. apply step_pushn.
eapply reduceS. apply step_swap.
eapply reduceS. apply step_op.
eapply reduceS. apply step_app.
eapply reduceS. apply step_pushv. reflexivity.
eapply reduceS. apply step_seltrue. red; intro. inversion H.
eapply reduceS. eapply step_pushc.
eapply reduceS. apply step_pushc.
eapply reduceS. apply step_pushv. reflexivity.
eapply reduceS. eapply step_pushn.
eapply reduceS. apply step_op.
eapply reduceS. apply step_app.
eapply reduceS. apply step_pushv. reflexivity.
eapply reduceS. apply step_pushv. reflexivity.
eapply reduceS. apply step_rapp.
eapply reduceS. apply step_pushc.
eapply reduceS. apply step_ret.
eapply reduceS. apply step_ret.
eapply reduceS. apply step_app.
eapply reduceS. eapply step_pushc.
eapply reduceS. apply step_pushv. reflexivity.
eapply reduceS. apply step_pushv. reflexivity.

eapply reduce_compose.
simpl in IHm |- *.
repeat rewrite <- minus_n_O. 
apply (IHm T).
eapply reduceS. apply step_app.
eapply reduceS. apply step_pushv. reflexivity.
eapply reduceS. apply step_pushv. reflexivity.
eapply reduceS. apply step_app.
eapply reduceS. apply step_pushv. reflexivity.
eapply reduceS. apply step_ret.
eapply reduceS. apply step_ret.
eapply reduceS. apply step_ret.
eapply reduceS. apply step_join.
eapply reduceS. apply step_ret.
eapply reduceS. apply step_ret.

apply reduce0.
Defined.

Lemma applyn_opt_terminate: forall T f n v e' e'' c e s d,
  f = id_fun e''
  -> terminate (App :: c, e,
    v :: VC [VN n , VRC [f] (compileV(applyn_fFn nil T)++[Ret]) , f] (compileE(applyn_fFnv nil T)++[Ret]) :: s, d)
  -> terminate (App :: c, e, v :: VC (VN n :: f :: e') [PushV 0, Ret] :: s, d).

intros.
eapply terminate_step. apply step_app.
eapply terminate_step. apply step_pushv. reflexivity.
eapply terminate_step. apply step_ret.

eapply reduce_terminate.
apply applyn_fFnv_reduce. 
rewrite H in H0.
apply H0.
Qed.

Lemma Sem_TLAM: forall E T1 T2 (tc: Exp (T1::E) T2),
  SemVal (TLAM tc) = sem_curry _ _ _ (SemExp tc).
auto.
Qed.

Lemma Sem_TVAL: forall E T (tv: Value E T),
  SemExp (TVAL tv) = sem_op _ eta (SemVal tv).
auto.
Qed.

Lemma applyn_opt_rel_ge: forall E T,
  (applyn_opt E T) >P> sem_op _ eta (SemVal (applyn E T)).

intros. red.
eapply ideal_closure_closed.
instantiate (1:= lub (sem_precomp_cha (sem_op_cha eta (SemValChain (applyn nil T))) (@K (SemEnv E) DOne tt))).
apply (@fcont_le_intro (SemEnv E) (DL (SemType ((T --> T) --> Int --> T --> T)))). intro.
rewrite sem_op_simpl, fcont_comp_simpl.
rewrite applyn_sem_simpl. rewrite <- (proj1 SemChain_sound).
repeat rewrite fcont_lub_simpl. rewrite lub_comp_eq_cont.
apply lub_le_compat.
apply fmon_le_intro. intros.
auto.

apply (proj1 (ideal_closure_closed (erelGeC (CC (applyn_opt E T))))).
intros. apply ideal_closure_sub.

do 4 red. intros.
rename e into pe, de into dpe, d into dapp, e0 into e, d0 into d.
unfold applyn_opt, ComposeApp.
repeat rewrite <- app_comm_cons.
eapply terminate_step. apply step_pushc.
unfold id_code, id_src, compileV. simpl app at 1.
eapply terminate_step. apply step_pushc.
eapply terminate_step. apply step_app.
eapply terminate_step. apply step_pushc.
eapply terminate_step. apply step_ret.

apply H1.
do 6 red. fold erelGeV. intros.
rename e0 into e_nil, de into de_nil, v into f, dv into df, d0 into dappf, e1 into e0, d1 into d0.
rewrite K_simpl in H4.
rewrite erelGeV_fold in H2.

repeat rewrite <- app_comm_cons.
eapply terminate_step.
apply step_app.
eapply terminate_step.
apply step_pushv. reflexivity.
eapply terminate_step.
apply step_pushv. reflexivity.
destruct (cesd_val_eq_dec f (id_fun (pe++e))).

Focus 2.
(* 
   case : f <> id_fun  
*)
eapply terminate_step.
apply step_eqfalse.
red; intros; apply H6; symmetry; auto.
eapply terminate_step.
apply step_selfalse.

unfold applyn_code, applyn, compileV; fold compileV.
repeat rewrite <- app_comm_cons.
eapply terminate_step.
apply step_pushc.
simpl app at 1.
eapply terminate_step.
apply step_pushv. reflexivity.
eapply terminate_step.
apply step_app.

unfold applyn_f, compileV; fold compileV.
repeat rewrite <- app_comm_cons.
eapply terminate_step.
apply step_pushrc.
eapply terminate_step.
apply step_ret.
eapply terminate_step.
apply step_join.
eapply terminate_step.
apply step_ret.

apply H5.
do 2 red. rewrite erelGeV_fold.
replace (compileV (applyn_fFn nil T)) with (compileE (TVAL (applyn_fFn nil T))).
eapply erelGeV_ge. Focus 2.
replace (f :: f :: VC (pe ++ e) [PushV 0, Ret] :: pe ++ e) with
  ([f]++(f :: VC (pe ++ e) [PushV 0, Ret] :: pe ++ e)).
eapply (@fix_correctness_ge i [T-->T] Int (T-->T)).
instantiate (1:= (tt, df)).
split. apply H2. apply erelGeENil_nil.
auto.

Focus 3. auto. Focus 2.
rewrite sem_precomp_cha_eval, sem_op_cha_eval, eta_val in H0.
hreplace H0 with (DLleVal_eq2 H0).
apply DLleVal_eq. eapply DLle_trans. apply H4.
apply Ole_to_DLle.
eapply Ole_trans. apply H0.
rewrite fcontit_simpl.

unfold applyn, SemValChain at 1; fold SemValChain SemEnv.
rewrite (fcont_eq_elim (fcont_eq_elim 
  (sem_curry_cha_eval (sem_op_cha eta (SemValChain (applyn_f nil T))) i) tt) df).
auto.

(* 
   case : f = id_fun 
*)
eapply terminate_step. apply step_eqtrue. apply symmetry. assumption. 
eapply terminate_step. apply step_seltrue. red; intros. inversion H7.
eapply terminate_step. apply step_pushc.
eapply terminate_step. apply step_join.
eapply terminate_step. apply step_ret.

apply H5.
do 4 red; intros. rewrite K_simpl in H9.
rename v into n, dv into dn, e1 into e1_nil, de into de1_nil, d1 into dappfn.
red; intros.

repeat rewrite <- app_comm_cons.
eapply terminate_step. apply step_app. unfold id_code.
eapply terminate_step. apply step_pushc.
eapply terminate_step. apply step_ret.

apply H10.
do 5 red; fold erelGeV; intros. rewrite K_simpl in H13.
rename e2 into e2_nil, de into de2_nil, d2 into dappfnv.
red; intros.

simpl in H7. rewrite H7.
simpl. apply applyn_opt_terminate with T (pe++e).
assumption.
eapply terminate_step.
apply step_app.
destruct compiler_correctness_ge_chain. clear H15.
hreplace H16 with (H16 _ _ (applyn_fFnv nil T) i).

hreplace H16 with (erelGeC_general H16).
rewrite (@app_nil_end ValueM [_,_,_,_]).
eapply (H16 _ (tt, df, SemValChain (applyn_f nil T) i (tt,df), dn, dv)).
split. apply H11. split. 
rewrite FST_simpl, Fst_simpl; rewrite SND_simpl, Snd_simpl. simpl. auto.
split. Focus 2. split. apply H2. apply erelGeENil_nil.
repeat rewrite FST_simpl, Fst_simpl; rewrite SND_simpl, Snd_simpl.

unfold applyn_f.
replace (compileV (applyn_fFn nil T)) with (compileE (TVAL (applyn_fFn nil T))).
rewrite (app_nil_end [f]).
apply (@fix_correctness_ge i [T-->T] Int (T-->T)
  (TVAL (applyn_fFn nil T)) [f] (tt,df)).
split. apply H2. apply erelGeENil_nil.
auto.

instantiate (1:= dappfnv).
rewrite sem_precomp_cha_eval, sem_op_cha_eval, eta_val in H0.
hreplace H0 with (DLleVal_eq2 H0).
hreplace H0 with (fcont_eq_elim H0 df).
unfold applyn, SemValChain in H0; fold SemValChain SemEnv in H0.
rewrite (fcont_eq_elim (fcont_eq_elim 
  (sem_curry_cha_eval (sem_op_cha eta (SemValChain (applyn_f nil T))) i) tt) df) in H0.
fold SemEnv in H0. unfold sem_curry in H0. rewrite CURRY_simpl, Curry_simpl in H0.
rewrite sem_op_cha_eval, eta_val in H0.
assert (H15 := DLleVal_eq2 (Oeq_trans H4 H0)).
assert (SemValChain (applyn_f nil T) i (tt, df) <= SemValChain (applyn_f nil T) (S i) (tt, df)).
apply (fmonotonic (SemValChain (applyn_f nil T))). auto.
hreplace H17 with (Ole_trans (proj1 H15) H17).

unfold applyn_f, SemValChain in H17; fold SemValChain SemEnv in H17.
rewrite sem_fix_cha_ind_S in H17.

replace (sem_fix_cha (sem_op_cha eta (SemValChain (applyn_fFn nil T))) i (tt, df)) with
  (SemValChain (applyn_f nil T) i (tt, df)) in H17.
hreplace H17 with (fcont_le_elim H17 dn).
repeat rewrite curry_simpl in H17.
rewrite sem_op_cha_eval, eta_val in H17.
assert (H18 := Ole_trans (proj1 H9) H17).
hreplace H18 with (DLleVal_eq H18).
hreplace H18 with (fcont_le_elim H18 dv).
eapply Ole_trans. apply H13.
eapply Ole_trans. apply H18.

unfold applyn_fFn, SemValChain at 1; fold SemExpChain SemEnv.
rewrite (sem_curry_cha_eval (SemExpChain (applyn_fFnv nil T)) i).
unfold sem_curry. rewrite CURRY_simpl, Curry_simpl. apply Ole_refl. auto.

red; intros.
eapply terminate_step. apply step_ret.
apply H14.
assumption.
Qed.

Fixpoint donothing_code k :=
  match k with
    | 0 => nil
    | S k' => DoNothing :: donothing_code k'
  end.

Definition div_code := [PushRC [PushV 1, PushV 0, App], PushN 0, App].

Lemma donothing_code_does_nothing: forall k c e s d,
  kstepreduce k (donothing_code k++c,e,s,d) (c,e,s,d).
induction k. intros; simpl. apply kstep0.
intros; simpl. eapply kstepS. apply step_donothing. apply IHk.
Qed.

Lemma id_fun_rel: forall k T e (df: SemType (T-->T)) x dx,
  (k > 3)%nat -> k:- id_fun e <v< df -> k:- x <v< dx ->
  exists dfx, df dx == Val dfx /\ k-3:- x <v< dfx.
intros kk T e df x dx H.
set (k:= kk-4). replace kk with (4+k). Focus 2. replace k with (kk-4). omega. auto.
intros.
assert (4+k <= 4+k)%nat. omega.
assert (H3 := 
  H0 (4+k) x dx H2 H1 (4+k) nil tt H2 (irelLeENil_nil (4+k)) 
  (4+k) (donothing_code k++EqHalt::div_code) nil [x] nil
); fold irelLeV in H3.
rewrite K_simpl in H3.

assert(stepsfor (4+k) ([App] ++(donothing_code k)++EqHalt :: div_code, nil++nil, [x, id_fun e] ++ [x], nil)->False).
intros.
assert (stepsfor 1 (EqHalt :: div_code, nil, [x, x], nil)).
eapply kstepreduce_stepsfor.
Focus 2. apply H4. instantiate (1:= 3+k).
eapply kstepreduce_compose. simpl.
eapply kstepS. apply step_app.
eapply kstepS. apply step_pushv. reflexivity.
eapply kstepS. apply step_ret. apply kstep0.
apply donothing_code_does_nothing. omega.
do 3 destruct H5. replace x0 with (S (x0-1)) in H6. inversion H6.
inversion H8. apply H19. auto. omega.

destruct (DL_bot_dec (df dx)).
absurd False; auto. apply H4. apply H3.
red; intros. do 2 destruct H6. 
absurd False; auto. apply (@DL_bot_neq _ x0).
rewrite <- H5. apply H6. omega. omega.

destruct H5. exists x0.
split. auto. replace (4+k-3) with (1 + k).
apply NNPP. red; intro.
apply H4. apply H3. red; intros.
destruct (lt_ge_dec i (k+1)).
exists k. eexists. split. omega. apply donothing_code_does_nothing.
exists (k+4). eexists. split. omega.
eapply kstepreduce_compose. apply donothing_code_does_nothing.
eapply kstepS. apply step_eqhalt.
red; intro. rewrite H10 in H7. apply H6.
do 2 destruct H7.
apply irelLeV_decr with i. omega.
apply irelLeV_le with x1.
rewrite H5 in H7. apply DLleVal_eq. apply H7.
assumption.

unfold div_code. simpl.
eapply kstepS. apply step_pushrc.
eapply kstepS. apply step_pushn.
eapply kstepS. apply step_rapp.
apply kstep0.
omega. omega. omega.
Qed.

Lemma id_fun_is_identity: forall k e (df: SemType (Int-->Int)) x dx,
  (k > 3)%nat -> k:- id_fun e <v< df -> k:- x <v< dx ->
  df dx == Val dx.

intros.
destruct (id_fun_rel H H0 H1) as [dn [H2 H3]].
rewrite H2. simpl in H1, H3. rewrite H1 in H3. inversion H3.
auto.
Qed.

Lemma applyn_f_sem_unfold: forall T df dn,
  SemVal (applyn_f nil T) (tt, df) dn == 
  Val (SemVal (applyn_fFn nil T) (tt, df, SemVal (applyn_f nil T) (tt, df), dn)).

intros.
unfold applyn_f, SemVal at 1. fold SemVal.
eapply Oeq_trans.
apply fcont_eq_elim.
rewrite sem_fix_eval.
rewrite fcont_comp_simpl.
rewrite FIXP_eq.
reflexivity.
repeat rewrite curry_simpl.
repeat rewrite fcont_comp_simpl.
repeat rewrite eta_val.
auto.
Qed.

Lemma applyn_fFn_eval_0: forall T (df:SemType (T-->T)) dv,
  SemVal (applyn_fFn nil T) (tt, df, SemVal (applyn_f nil T) (tt, df), 0) dv == Val dv.

intros.
unfold applyn_fFn, applyn_fFnv, SemVal; fold SemExp SemVal SemType.
unfold VTm, VTmVar, SemVar.
unfold sem_curry. rewrite CURRY_simpl, Curry_simpl.
rewrite sem_op_simpl.
rewrite (sem_op_simpl (A:= Discrete bool) (B:=DL (Discrete bool))).
rewrite (sem_app_val_simpl (E:= [T,Int,Int-->T-->T,T-->T]) (A:= Bool) (B:= T)).
rewrite fcont_comp_simpl.
rewrite PROD_fun_simpl, PAIR_simpl, Pair_simpl.
unfold EV. fold SemType. rewrite fcont_comp2_simpl.
rewrite FST_simpl, Fst_simpl. unfold fst.
rewrite SND_simpl, Snd_simpl. unfold snd.
rewrite AP_simpl.
rewrite CURRY_simpl, Curry_simpl.
rewrite sem_op3_eval at 1.
rewrite choose_f_simpl.
rewrite sem_op_eval at 1.
unfold SemVal at 1. unfold SemVar. rewrite fcont_comp_simpl.
rewrite eta_val.
rewrite FST_simpl, Fst_simpl. unfold fst.
rewrite (SND_simpl (D1:=SemEnv [Int,Int-->T-->T,T-->T]) (D2:= SemType T)). rewrite Snd_simpl. unfold snd.
auto.
Qed.

Lemma applyn_fFn_eval_S: forall T (df:SemType (T-->T)) dn dv dval,
  SemVal (applyn_fFn nil T) (tt, df, SemVal (applyn_f nil T) (tt, df), dn) dv == Val dval ->
  SemVal (applyn_fFn nil T) (tt, df, SemVal (applyn_f nil T) (tt, df), S dn) dv == df dval.

intros.
unfold applyn_f.
introvar appnfFnnilT (applyn_fFn nil T) HappnfFn.
rewrite <- HappnfFn at 2. rewrite <- HappnfFn in H.
unfold applyn_fFn, SemVal; fold SemExp SemVal SemType.
unfold sem_curry. rewrite CURRY_simpl, Curry_simpl.
unfold applyn_fFnv, SemExp; fold SemExp SemVal SemType.
rewrite sem_op_simpl.
rewrite (sem_op_simpl (A:= Discrete bool) (B:=DL (Discrete bool))).
rewrite (sem_app_val_simpl (E:= [T,Int,Int-->T-->T,T-->T]) (A:= Bool) (B:= T)).
rewrite fcont_comp_simpl.
rewrite PROD_fun_simpl, PAIR_simpl, Pair_simpl.
unfold EV. fold SemType. rewrite fcont_comp2_simpl.
rewrite FST_simpl, Fst_simpl. unfold fst.
rewrite SND_simpl, Snd_simpl. unfold snd.
rewrite AP_simpl.
rewrite CURRY_simpl, Curry_simpl.
rewrite sem_op3_eval at 1.
rewrite choose_t_simpl.

erewrite <- sem_app_simpl.
Focus 2. rewrite sem_op_eval, eta_val. apply Oeq_refl.
Focus 2.
  rewrite sem_op_simpl. 
  rewrite (sem_op_simpl (A:=Discrete nat) (B:=DL (Discrete nat))).
  rewrite (sem_app_val_simpl (E:= [Bool,T,Int,Int-->T-->T,T-->T]) (A:= Int) (B:= T-->T)).
  rewrite fcont_comp_simpl.
  rewrite PROD_fun_simpl, PAIR_simpl, Pair_simpl.
  unfold EV. fold SemType. rewrite fcont_comp2_simpl.
  rewrite FST_simpl, Fst_simpl. unfold fst.
  rewrite SND_simpl, Snd_simpl. unfold snd.
  rewrite AP_simpl.
  rewrite CURRY_simpl, Curry_simpl.
  rewrite sem_op_simpl. rewrite sem_op_simpl.
  rewrite (sem_app_val_simpl (A:= Int) (B:= T-->T)).
  rewrite fcont_comp_simpl.
  rewrite PROD_fun_simpl, PAIR_simpl, Pair_simpl.
  unfold VTm, VTmVar. unfold SemVal. fold SemVal SemExp. unfold SemVar.
  repeat rewrite fcont_comp_simpl.
  repeat rewrite FST_simpl, Fst_simpl. unfold fst.
  rewrite (SND_simpl (D1:=SemEnv [T-->T]) (D2:= SemType (Int-->T-->T))). rewrite Snd_simpl. unfold snd.  
  rewrite (SND_simpl (D1:=SemEnv [Bool,T,Int,Int-->T-->T,T-->T]) (D2:= SemType Int)). rewrite Snd_simpl. unfold snd.
  rewrite sem_op2_eval. rewrite SimpleOp2_simpl. simpl OpSem. rewrite <- minus_n_O.
  unfold EV. fold SemType. rewrite fcont_comp2_simpl.
  rewrite FST_simpl, Fst_simpl. unfold fst.
  rewrite SND_simpl, Snd_simpl. unfold snd.
  rewrite AP_simpl.

  instantiate (1:= SemVal appnfFnnilT (tt, df, SemVal (applyn_f nil T) (tt, df), dn)).
  rewrite sem_fix_eval.
  rewrite fcont_comp_simpl.
  rewrite FIXP_eq.
  repeat rewrite curry_simpl.
  rewrite sem_op_eval.
  rewrite eta_val.
  apply (Val_stable (D:= SemType (T-->T))).
  apply fcont_eq_compat.
  apply PAIR_eq_compat. apply PAIR_eq_compat. auto. 
  unfold applyn_f, SemVal; fold SemVal SemExp.
  rewrite sem_fix_eval. rewrite fcont_comp_simpl. rewrite HappnfFn. reflexivity.
  auto.

unfold sem_curry. rewrite CURRY_simpl, Curry_simpl.
erewrite <- sem_app_simpl.
apply Oeq_refl.
  
rewrite sem_op_eval, eta_val.
apply (Val_stable (D:= SemType (T-->T))).
apply (@fcont_eq_intro (SemType T) (DL (SemType T))). intros.
rewrite CURRY_simpl, Curry_simpl.
rewrite sem_op_simpl. rewrite sem_op_simpl. rewrite sem_app_val_simpl.
rewrite fcont_comp_simpl.
rewrite PROD_fun_simpl, PAIR_simpl, Pair_simpl.
unfold EV. fold SemType. rewrite fcont_comp2_simpl.
rewrite FST_simpl, Fst_simpl. unfold fst.
rewrite SND_simpl, Snd_simpl. unfold snd.
rewrite AP_simpl.
unfold VTm, VTmVar. unfold SemVal. fold SemVal SemExp. unfold SemVar.
repeat rewrite fcont_comp_simpl.
repeat rewrite FST_simpl, Fst_simpl. unfold fst.
rewrite (SND_simpl (D1:=SemEnv [T-->T,Bool,T,Int,Int-->T-->T,T-->T]) (D2:= SemType T)). rewrite Snd_simpl. unfold snd.  
apply Oeq_refl.

apply Oeq_sym.
rewrite sem_op_simpl. rewrite sem_op_simpl.
rewrite sem_app_val_simpl.
rewrite fcont_comp_simpl.
rewrite PROD_fun_simpl, PAIR_simpl, Pair_simpl.
unfold EV. fold SemType. rewrite fcont_comp2_simpl.
rewrite FST_simpl, Fst_simpl. unfold fst.
rewrite SND_simpl, Snd_simpl. unfold snd.
rewrite AP_simpl.
unfold VTm, VTmVar. unfold SemVal. fold SemVal SemExp. unfold SemVar.
repeat rewrite fcont_comp_simpl.
repeat rewrite FST_simpl, Fst_simpl. unfold fst.
rewrite (SND_simpl (D1:=SemEnv [Bool,T,Int,Int-->T-->T,T-->T]) (D2:= SemType (T-->T))). rewrite Snd_simpl. unfold snd.
rewrite (SND_simpl (D1:=SemEnv [Int,Int-->T-->T,T-->T]) (D2:= SemType T)). rewrite Snd_simpl. unfold snd.  
rewrite H.
apply Oeq_refl.
Qed.

Lemma applyn_fFn_id: forall T (df:SemType (T-->T)) dv dn, df dv == Val dv ->
  SemVal (applyn_fFn nil T) (tt, df, SemVal (applyn_f nil T) (tt, df), dn) dv == Val dv.

intros. simpl in dn. 
induction dn.

rewrite applyn_fFn_eval_0. auto.
rewrite <- H.
apply applyn_fFn_eval_S. auto.
Qed.

Lemma applyn_opt_rel_le E:
  (applyn_opt E Int) <P< sem_op _ eta (SemVal (applyn E Int)).
Proof.
set (T := Int).

do 4 red; unfold CC; intros.
rename e into pe, de into dpe, e0 into e.

unfold applyn_opt, ComposeApp. repeat rewrite app_ass.
eapply stepsfor_step. apply step_pushc. unfold id_code, id_src, compileV. 
eapply stepsfor_step. apply step_pushc.
eapply stepsfor_step. apply step_app.
eapply stepsfor_step. apply step_pushc.
eapply stepsfor_step. apply step_ret.

apply H1 with k1.
exists (SemVal (applyn nil T) tt).
split. rewrite sem_op_simpl. repeat rewrite fcont_comp_simpl.
rewrite (@applyn_sem_simpl E T dpe). auto.
red; fold irelLeV; red; intros.
red; intros.
rename v into f, dv into df, e0 into e_nil, de into de_nil.
red; intros. rewrite K_simpl in H8.

repeat rewrite <- app_comm_cons.
eapply stepsfor_step. apply step_app.
eapply stepsfor_step. apply step_pushv. reflexivity.
eapply stepsfor_step. apply step_pushv. reflexivity.

destruct (cesd_val_eq_dec f (id_fun (pe++e))).

Focus 2.
(* 
   case : f <> id_fun 
*)
eapply stepsfor_step.
apply step_eqfalse.
red; intros. apply H11. apply symmetry. assumption.
eapply stepsfor_step.
apply step_selfalse.

unfold applyn_code, applyn, compileV; fold compileV.
repeat rewrite <- app_comm_cons.
eapply stepsfor_step.
apply step_pushc.
eapply stepsfor_step.
apply step_pushv. reflexivity.
eapply stepsfor_step.
apply step_app.

unfold applyn_f, compileV; fold compileV.
repeat rewrite <- app_comm_cons.
eapply stepsfor_step.
apply step_pushrc.
eapply stepsfor_step.
apply step_ret.
eapply stepsfor_step.
apply step_join.
eapply stepsfor_step.
apply step_ret.
apply H8 with k4.

exists (SemVal (TFIX (TVAL (applyn_fFn nil T))) (tt,df)).
split. unfold applyn, applyn_f. unfold SemVal; fold SemVal.
repeat rewrite curry_simpl.
repeat rewrite fcont_comp_simpl. auto.
replace (f :: f :: VC (pe ++ e) (PushV (idxVar (ZVAR E T)) :: nil ++ [Ret]) :: pe ++ e)
  with ([f]++(f :: VC (pe ++ e) (PushV (idxVar (ZVAR E T)) :: nil ++ [Ret]) :: pe ++ e)).
apply fix_correctness_le.
split. apply irelLeV_decr with k2. omega. apply H5.
apply irelLeENil_nil.

auto. omega. omega.

(* 
   case : f = id_fun 
*)
eapply stepsfor_step. apply step_eqtrue. apply symmetry. assumption.
eapply stepsfor_step. apply step_seltrue. red; intros. inversion H12.
eapply stepsfor_step. apply step_pushc.
eapply stepsfor_step. apply step_join.
eapply stepsfor_step. apply step_ret.

apply H8 with k4.
unfold applyn. unfold SemVal; fold SemVal.
unfold sem_curry. rewrite CURRY_simpl, Curry_simpl.
exists (SemVal (applyn_f nil T) (tt, df)).
split. auto.
do 3 red; fold irelLeV; intros.
rename e1 into e1_nil, de into de1_nil, v into n, dv into dn.
red; intros. 

repeat rewrite <- app_comm_cons.
eapply stepsfor_step. apply step_app. simpl.
eapply stepsfor_step. apply step_pushc.
eapply stepsfor_step. apply step_ret.

rewrite K_simpl in H16.
apply H16 with k7.
exists (SemVal (applyn_fFn nil T) (tt, df, SemVal(applyn_f nil T) (tt,df), dn)).
split.

unfold applyn_f. unfold SemVal; fold SemVal.
apply Oeq_sym. eapply Oeq_trans.
rewrite sem_fix_eval, fcont_comp_simpl, FIXP_eq.
repeat rewrite curry_simpl.
rewrite sem_op_simpl. rewrite fcont_comp_simpl, eta_val. apply Oeq_refl.
auto.

do 3 red; fold irelLeV; intros.
rename e2 into e2_nil, de into de2_nil.
simpl.

eapply stepsfor_step_strict. apply step_app.
eapply stepsfor_step_strict. apply step_pushv. reflexivity.
eapply stepsfor_step_strict. apply step_ret.

destruct e2_nil. Focus 2. inversion H22. simpl. 
destruct (le_gt_dec k10 3).
apply stepsfor_decr with 0. omega. apply stepsfor_zero.

apply H23 with k10.
rewrite K_simpl.
assert (k10:- id_fun (pe++e) <v< df).
rewrite <- H11. apply irelLeV_decr with k2. omega. assumption.
assert (k10:- v <v< dv). apply irelLeV_decr with k8. omega. assumption.
assert (H28 := id_fun_is_identity g H26 H27).

exists dv.
split.
apply Oeq_sym.
apply (@applyn_fFn_id Int _ _ dn H28).
assumption.
omega. omega. omega. omega. omega. omega. omega. omega.
Qed.

Corollary applyn_opt_rel E:
  (applyn_opt E Int) =P= sem_op _ eta (SemVal (applyn E Int)).

intros; split.
apply applyn_opt_rel_le.
apply applyn_opt_rel_ge.
Qed.

Corollary applyn_opt_eq E:
  E |- (applyn_opt E Int) ~P~ (applyn_code E Int) ::: ((Int-->Int)-->Int-->Int-->Int).

intros.
exists (sem_op _ eta (SemVal (applyn E Int))).
split.
apply applyn_opt_rel.
unfold applyn_code.
apply compiler_correctness.
Qed.

Lemma run_applyn_opt: forall E cesd k n,
  kstepreduce 22 
  (mergeComp (CC(ComposeApp (ComposeApp (ComposeApp (applyn_opt E Int) (id_code E Int)) [PushN k]) [PushN n])) cesd)
  (mergeComp (VToC (VN n)) cesd).

intros.
destruct cesd as [[[c e] s ]d]. simpl. 
eapply kstepS. apply step_pushc. 
eapply kstepS. apply step_pushc.
eapply kstepS. apply step_app.
eapply kstepS. apply step_pushc.
eapply kstepS. apply step_ret.
eapply kstepS. apply step_pushc.
eapply kstepS. apply step_app.
eapply kstepS. apply step_pushv. reflexivity.
eapply kstepS. apply step_pushv. reflexivity.
eapply kstepS. apply step_eqtrue; auto.
eapply kstepS. apply step_seltrue.
red; intros.
inversion H.
eapply kstepS. apply step_pushc.
eapply kstepS. apply step_join.
eapply kstepS. apply step_ret.
eapply kstepS. apply step_pushn.
eapply kstepS. apply step_app.
eapply kstepS. apply step_pushc.
eapply kstepS. apply step_ret.
eapply kstepS. apply step_pushn.
eapply kstepS. apply step_app.
eapply kstepS. apply step_pushv. reflexivity.
eapply kstepS. apply step_ret.
apply kstep0.
Qed.

Lemma run_apply_n: forall E cesd n,
  kstepreduce 59 
  (mergeComp (CC(ComposeApp (ComposeApp (ComposeApp (applyn_code E Int) (id_code E Int)) [PushN 1]) [PushN n])) cesd)
  (mergeComp (VToC (VN n)) cesd).

intros.
destruct cesd as [[[c e] s ]d]. simpl. 
eapply kstepS. apply step_pushc.
eapply kstepS. apply step_pushc.
eapply kstepS. apply step_app.
eapply kstepS. apply step_pushrc.
eapply kstepS. apply step_ret.
eapply kstepS. apply step_pushn.
eapply kstepS. apply step_rapp.
eapply kstepS. apply step_pushc.
eapply kstepS. apply step_ret.
eapply kstepS. apply step_pushn.
eapply kstepS. apply step_app.
eapply kstepS. apply step_pushc.
eapply kstepS. apply step_pushv. reflexivity.
eapply kstepS. apply step_pushn.
eapply kstepS. apply step_swap.
eapply kstepS. apply step_op.
eapply kstepS. apply step_app. 
eapply kstepS. apply step_pushv. reflexivity.
eapply kstepS. apply step_seltrue; red; intros; inversion H.
eapply kstepS. apply step_pushc.
eapply kstepS. apply step_pushc.
eapply kstepS. apply step_pushv. reflexivity.
eapply kstepS. apply step_pushn.
eapply kstepS. apply step_op.
eapply kstepS. apply step_app.
eapply kstepS. apply step_pushv. reflexivity.
eapply kstepS. apply step_pushv. reflexivity.
eapply kstepS. apply step_rapp.
eapply kstepS. apply step_pushc.
eapply kstepS. apply step_ret.
eapply kstepS. apply step_ret.
eapply kstepS. apply step_app.
eapply kstepS. apply step_pushc.
eapply kstepS. apply step_pushv. reflexivity.
eapply kstepS. apply step_pushv. reflexivity.
eapply kstepS. apply step_app.
eapply kstepS. apply step_pushc.
eapply kstepS. apply step_pushv. reflexivity.
eapply kstepS. apply step_pushn.
eapply kstepS. apply step_swap.
eapply kstepS. apply step_op.
eapply kstepS. apply step_app.
eapply kstepS. apply step_pushv. reflexivity.
eapply kstepS. apply step_selfalse.
eapply kstepS. apply step_pushv. reflexivity.
eapply kstepS. apply step_join.
eapply kstepS. apply step_ret.
eapply kstepS. apply step_ret.
eapply kstepS. apply step_app.
eapply kstepS. apply step_pushv. reflexivity.
eapply kstepS. apply step_pushv. reflexivity.
eapply kstepS. apply step_app.
eapply kstepS. apply step_pushv. reflexivity.
eapply kstepS. apply step_ret.
eapply kstepS. apply step_ret.
eapply kstepS. apply step_ret.
eapply kstepS. apply step_join.
eapply kstepS. apply step_ret.
eapply kstepS. apply step_ret.
apply kstep0.
Qed.

